#! /usr/bin/env python3
"""
Test functions in munging.py
"""

from typing import Callable
from datetime import datetime
import time

import dawgdad as dd
import pandas as pd
import numpy as np

output_url = 'test_munging.html'
header_title = 'test_munging'
header_id = 'test-munging'
file_name = 'test_munging.csv'


def main():
    start_time = time.time()
    pd.options.display.width = 120
    pd.options.display.max_columns = 100
    pd.options.display.max_rows = 100
    original_stdout = dd.html_begin(
        output_url=output_url,
        header_title=header_title,
        header_id=header_id
    )
    print('--------------------------')
    print('test dataframe_info')
    print('test example 1')
    my_file = 'myfile.csv'
    df = dd.read_file(file_name=my_file)
    df = dd.dataframe_info(
        df=df,
        file_in=my_file
    )
    print('--------------------------')
    print('test dataframe_info')
    print('test example 2')
    df = dd.create_dataframe()
    df = dd.dataframe_info(
        df=df,
        file_in='df'
    )
    print('--------------------------')
    print('test find_bool_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_bool = dd.find_bool_columns(df=df)
    print(columns_bool)
    print('--------------------------')
    print('test find_category_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_category = dd.find_category_columns(df=df)
    print(columns_category)
    print('--------------------------')
    print('test find_datetime_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_datetime = dd.find_datetime_columns(df=df)
    print(columns_datetime)
    print('--------------------------')
    print('test find_float_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_float = dd.find_float_columns(df=df)
    print(columns_float)
    print('--------------------------')
    print('test find_int_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_int = dd.find_int_columns(df=df)
    print(columns_int)
    print('--------------------------')
    print('test find_int_float_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_int_float = dd.find_int_float_columns(df=df)
    print(columns_int_float)
    print('--------------------------')
    print('test find_object_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_object = dd.find_object_columns(df=df)
    print(columns_object)
    print('--------------------------')
    print('test find_timedelta_columns')
    print('test example')
    df = dd.create_dataframe()
    columns_timedelta = dd.find_timedelta_columns(df=df)
    print(columns_timedelta)
    print('--------------------------')
    print('test number_empty_cells_in_columns')
    print('test example')
    df = pd.DataFrame({
        'X': [25.0, 24.0, 35.5, np.nan, 23.1],
        'Y': [27, 24, np.nan, 23, np.nan],
        'Z': ['a', 'b', np.nan, 'd', 'e']
    })
    empty_cells = dd.number_empty_cells_in_columns(df=df)
    print(empty_cells)
    print('--------------------------')
    print('test process_columns')
    print('test example')
    df = dd.create_dataframe()
    df, columns_in_count, columns_non_empty_count, columns_empty_count,\
        columns_empty_list, columns_non_empty_list, columns_bool_list,\
        columns_bool_count, columns_float_list, columns_float_count,\
        columns_integer_list, columns_integer_count,\
        columns_datetime_list, columns_datetime_count,\
        columns_object_list, columns_object_count, columns_category_list,\
        columns_category_count, columns_timedelta_list,\
        columns_timedelta_count = dd.process_columns(df=df)
    print('columns_in_count       :', columns_in_count)
    print('columns_non_empty_count:', columns_non_empty_count)
    print('columns_empty_count    :', columns_empty_count)
    print('columns_empty_list     :', columns_empty_list)
    print('columns_non_empty_list :', columns_non_empty_list)
    print('columns_bool_list      :', columns_bool_list)
    print('columns_bool_count     :', columns_bool_count)
    print('columns_float_list     :', columns_float_list)
    print('columns_float_count    :', columns_float_count)
    print('columns_integer_list   :', columns_integer_list)
    print('columns_integer_count  :', columns_integer_count)
    print('columns_datetime_list  :', columns_datetime_list)
    print('columns_datetime_count :', columns_datetime_count)
    print('columns_object_list    :', columns_object_list)
    print('columns_object_count   :', columns_object_count)
    print('columns_category_list  :', columns_category_list)
    print('columns_category_count :', columns_category_count)
    print('columns_timedelta_list :', columns_timedelta_list)
    print('columns_timedelta_count:', columns_timedelta_count)
    print('--------------------------')
    print('test process_rows')
    print('test example')
    df = dd.create_dataframe()
    df, rows_in_count, rows_out_count, rows_empty_count = dd.process_rows(
        df=df
    )
    print('rows_in_count   :', rows_in_count)
    print('rows_out_count  :', rows_out_count)
    print('rows_empty_count:', rows_empty_count)
    print('--------------------------')
    print('test save_file example 1')
    print('test example')
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name='x_y.csv'
    )
    print('--------------------------')
    print('test save_file example 2')
    print('test example')
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name='x_y.csv',
        index=True,
        index_label='myindex'
    )
    print('--------------------------')
    print('test save_file example 3')
    print('test example')
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name='x_y.xlsx'
    )
    print('--------------------------')
    print('test save_file example 4')
    print('test example')
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name='x_y.xlsx',
        index=True,
        index_label='myindex'
    )
    print('--------------------------')
    print('test byte_size')
    print('test example')
    df = dd.create_dataframe()
    print(
        dd.byte_size(
            num=df.memory_usage(index=True).sum()
        )
    )
    print('--------------------------')
    print('test read_file')
    print('test example 1')
    my_file = 'myfile.csv'
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name=my_file
    )
    df = dd.read_file(file_name=my_file)
    dd.dataframe_info(
        df=df,
        file_in=my_file)
    stop_time = time.time()
    print('--------------------------')
    print('test read_file')
    print('test example 2')
    file_name = 'myfile.csv'
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name=file_name
    )
    parse_dates = ['t', 'u']
    df = dd.read_file(
        file_name=file_name,
        parse_dates=parse_dates
    )
    dd.dataframe_info(
        df=df,
        file_in=my_file
    )
    print('--------------------------')
    print('test read_file')
    print('test example 3')
    file_name = 'myfile.csv'
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name=file_name
    )
    column_names_dict = {
        'a': 'A',
        'b': 'B',
        'c': 'C',
        'd': 'D',
        'i': 'I',
        'r': 'R',
        's': 'S',
        't': 'T',
        'u': 'U',
        'y': 'Y',
        'x': 'X',
        'z': 'Z'
    }
    index_columns = ['Y']
    parse_dates = ['t', 'u']
    time_delta_columns = ['D']
    category_columns = ['C']
    integer_columns = ['A', 'I']
    float_columns = ['X']
    boolean_columns = ['R']
    object_columns = ['Z']
    df = dd.read_file(
        file_name=file_name,
        column_names_dict=column_names_dict,
        index_columns=index_columns,
        date_parser=date_parser(),
        parse_dates=parse_dates,
        time_delta_columns=time_delta_columns,
        category_columns=category_columns,
        integer_columns=integer_columns
    )
    dd.dataframe_info(
        df=df,
        file_in=file_name
    )
    print('--------------------------')
    print('test read_file')
    print('test example 4')
    file_name = 'myfile.odd.
    df = dd.create_dataframe()
    dd.save_file(
        df=df,
        file_name=file_name
    )
    parse_dates = ['t', 'u']
    df = dd.read_file(
        file_name=file_name,
        parse_dates=parse_dates
    )
    dd.dataframe_info(
        df=df,
        file_in=file_name
    )
    stop_time = time.time()
    dd.page_break()
    dd.report_summary(
        start_time=start_time,
        stop_time=stop_time,
        read_file_names=file_name,
        save_file_names=file_name
    )
    dd.html_end(
        original_stdout=original_stdout,
        output_url=output_url
    )


def date_parser() -> Callable:
    """
    Date parser callable function

    Returns
    -------
    Parsed date and time.

    Example
    >>> date_parser=date_parser()
    """
    return lambda s: datetime.strptime(s, '%Y-%m-%d %H:%M:%S')


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
'''
Test help for dawgdad.

time -f '%e' ./dawgdad.help_test.py
./dawgdad.help_test.py

Typical input:
dd.stats.random_data
'''

import dawgdad as dd

output_url = 'dawgdad.help_test.html'
header_title = 'dawgdad.help'
header_id = 'dawgdad.help'


def main():
    input_value = eval(input(r'module.file.function name? > '))
    original_stdout = dd.html_begin(
        output_url=output_url,
        header_title=header_title,
        header_id=header_id
    )
    print('<pre style="white-space: pre-wrap;">')
    help(input_value)
    print('</pre>')
    dd.html_end(
        original_stdout=original_stdout,
        output_url=output_url
    )


if __name__ == '__main__':
    main()



import dawgdad as dd


def test_water_coffee_tea_milk():
    result = dd.water_coffee_tea_milk(
            mugs_coffee=0,
            cups_tea=0,
            mugs_tea=0
        )
    expected = (0, 0, 0, 0, 0, 0, 0, (0, 0, 0))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=0,
            cups_tea=0,
            mugs_tea=1
        )
    expected = (300, 0, 0, 0, 300, 0, 0, (0, 1, 42))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=0,
            cups_tea=1,
            mugs_tea=0
        )
    expected = (400, 0, 0, 400, 0, 0, 0, (0, 2, 16))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=0,
            cups_tea=1,
            mugs_tea=1
        )
    expected = (700, 0, 0, 400, 300, 0, 0, (0, 3, 58))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=1,
            cups_tea=0,
            mugs_tea=0
        )
    expected = (370, 220, 150, 0, 0, 150, 20, (0, 2, 5))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=1,
            cups_tea=0,
            mugs_tea=1
        )
    expected = (670, 220, 150, 0, 300, 150, 20, (0, 3, 47))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=1,
            cups_tea=1,
            mugs_tea=0
        )
    expected = (770, 220, 150, 400, 0, 150, 20, (0, 4, 21))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=1,
            cups_tea=1,
            mugs_tea=1
        )
    expected = (1070, 220, 150, 400, 300, 150, 20, (0, 6, 3))
    assert result == expected
    result = dd.water_coffee_tea_milk(
            mugs_coffee=2,
            cups_tea=0,
            mugs_tea=0
        )
    expected = (740, 440, 300, 0, 0, 300, 40, (0, 4, 11))
    assert result == expected



import dawgdad as dd


def test_list_empty_except_nan_worksheet_rows():
    pass


def test_list_empty_and_nan_worksheet_rows():
    pass


def test_list_duplicate_worksheet_rows():
    pass


def test_change_case_worksheet_columns():
    pass


def test_write_dataframe_to_worksheet():
    pass


def test_remove_empty_worksheet_rows():
    pass


def test_remove_worksheet_columns():
    pass


def test_list_nan_worksheet_rows():
    pass


def test_list_rows_with_content():
    pass


def test_validate_column_labels():
    pass


def test_number_non_empty_rows():
    pass


def test_remove_worksheet_rows():
    pass


def test_validate_sheet_names():
    pass


def test_autofit_column_width():
    pass


def test_unique_list_items():
    pass


def test_cell_fill_down():
    pass


def test_read_workbook():
    pass


def test_replace_text():
    pass


def test_exit_script():
    pass


def test_cell_style():
    pass



import dawgdad.msa as msa


def test_range_chart():
    pass
    # raise NotImplementedError()


def test_average_chart():
    pass
    # raise NotImplementedError()


def test_parallelism_plot():
    pass
    # raise NotImplementedError()


def test_main_effects_chart_anome():
    pass
    # raise NotImplementedError()


def test_mean_ranges_chart_anomr():
    pass
    # raise NotImplementedError()


def test_msa_results():
    pass
    # raise NotImplementedError()


def test_classification():
    pass
    # raise NotImplementedError()


def test_effective_resolution():
    pass
    # raise NotImplementedError()


def test_variance_components():
    pass
    # raise NotImplementedError()


def test_grr_gauge_rr_results():
    pass
    # raise NotImplementedError()


def test_interpret():
    pass
    # raise NotImplementedError()


def test_interpret_tables():
    pass
    # raise NotImplementedError()


def test_range_in_control():
    pass
    # raise NotImplementedError()


def test_range_out_of_control_reason():
    pass
    # raise NotImplementedError()


def test_average_in_control():
    pass
    # raise NotImplementedError()


def test_average_out_of_control_reason():
    pass
    # raise NotImplementedError()


def test_main_effects_in_control():
    pass
    # raise NotImplementedError()


def test_main_effects_out_of_control_reason():
    pass
    # raise NotImplementedError()


def test_mean_ranges_in_control():
    pass
    # raise NotImplementedError()


def test_mean_ranges_out_of_control_reason():
    pass
    # raise NotImplementedError()



#! /usr/bin/env python3
"""
Create prerender graphs for control chart tests
"""

from pathlib import Path
import time

from dawgdad.import control_charts as cc
import matplotlib.pyplot as plt
import dawgdad as dd
import pandas as pd


def main():
    df = pd.DataFrame({
        'Sample': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
        'X': [25.0, 24.0, 35.5, 22.4, 23.1, 13.9, 13.9, 10.0, 13.3, 10.0, 16.0,
              16.0, 16.0],
    }).set_index('Sample')
    X = cc.X(data=df[['X']])
    ax = X.ax()
    cc.draw_rules(
        cc=X,
        ax=ax
    )
    ax.figure.savefig(
        fname=f"{Path('../', 'tests', 'prerenders', 'X-rules')}.png",
        format='png'
    )


if __name__ == "__main__":
    main()



import dawgdad as dd


def test_plot_scatterleft_scatterright_x_y1_y2():
    pass


def test_plot_scatter_scatter_x1_x2_y1_y2():
    pass


def test_plot_lineleft_lineright_x_y1_y2():
    pass


def test_plot_barleft_lineright_x_y1_y2():
    pass


def test_plot_line_line_line_x_y1_y2_y3():
    pass


def test_plot_scatter_scatter_x_y1_y2():
    pass


def test_plot_scatter_line_x_y1_y2():
    pass


def test_plot_line_line_x_y1_y2():
    pass


def test_plot_horizontal_bars():
    pass


def test_plot_line_line_y1_y2():
    pass


def test_plot_vertical_bars():
    pass


def test_plot_stacked_bars():
    pass


def test_probability_plot():
    pass


def test_plot_scatter_x_y():
    pass


def test_plot_histogram():
    pass


def test_plot_scatter_y():
    pass


def test_plot_line_x_y():
    pass


def test_format_dates():
    pass


def test_plot_boxplot():
    pass


def test_plot_line_y():
    pass


def test_plot_pareto():
    pass


def test_style_graph():
    pass


def test_dd_to_dms():
    decdeg = [45.4250225, -75.6970594]
    result = dd.dd_to_dms(decdeg=decdeg)
    expected = [(45, 25, 30.081, 'N'), (75, 41, 49.41384, 'W')]
    assert result == expected
    decdeg = [48.858393, 2.257616]
    result = dd.dd_to_dms(decdeg=decdeg)
    expected = [(48, 51, 30.2148, 'N'), (2, 15, 27.4176, 'E')]
    assert result == expected
    decdeg = [-13.163194, -72.547842]
    result = dd.dd_to_dms(decdeg=decdeg)
    expected = [(13, 9, 47.4984, 'S'), (72, 32, 52.2312, 'W')]
    assert result == expected
    decdeg = [-33.8567433, 151.1784306]
    result = dd.dd_to_dms(decdeg=decdeg)
    expected = [(33, 51, 24.27588, 'S'), (151, 10, 42.35016, 'E')]
    assert result == expected

def test_dms_to_dd():
    dms = [(45, 25, 30.081, 'N'), (75, 41, 49.41384, 'W')]
    result = dd.dms_to_dd(dms=dms)
    expected = [45.4250225, -75.6970594]
    assert result == expected
    dms = [(48, 51, 30.2148, 'N'), (2, 15, 27.4176, 'E')]
    result = dd.dms_to_dd(dms=dms)
    expected = [48.858393, 2.257616]
    assert result == expected
    dms = [(13, 9, 47.4984, 'S'), (72, 32, 52.2312, 'W')]
    result = dd.dms_to_dd(dms=dms)
    expected = [-13.163194, -72.547842]
    assert result == expected
    dms = [(33, 51, 24.27588, 'S'), (151, 10, 42.35016, 'E')]
    result = dd.dms_to_dd(dms=dms)
    expected = [-33.8567433, 151.1784306]
    assert result == expected


def test_plot_pie():
    pass


def test_despine():
    pass


def test_qr_code():
    pass



#! /usr/bin/env python3
'''
Syntax for functions

- Parameters without a default argument must explicitly called
'''

from inspect import signature
from typing import List, Tuple, Union


def function_name_syntax(
    positional_only_parameters,
    /,
    positional_or_keyword_parameters,
    *,
    keyword_only_parameters
):
    pass


def function_name(
    a: float,
    /,
    b: float,
    c: float = 6,
    d: Union[float, None] = None,
    e: Union[float, int, None] = None,
    *,
    f: float,
    g: float = 4,
    h: Union[float, None] = None,
    i: Union[float, int, None] = None,
    j: Union[List[str], Tuple[float, int], float, int, None] = None,
):
    pass


def main():
    print(signature(function_name))
    for param in signature(function_name).parameters.values():
        print(param, param.kind.description)
    # function_name(1, 2, f=3)
    # function_name(1, b=2, f=3)
    # function_name(1, 2, 13, f=3)
    # function_name(1, 2, c=13, f=3)


if __name__ == '__main__':
    main()



"""
Process capability refers to the ability of a process to meet a performance
standard (specification). A process is capable if you have:
- Specifications are defined and attainable.
- Can measure sufficiently well.
- Samples are representative.
- Process variation is stable and predictable.
- Process is on target with minimum dispersion.
"""

import math

from scipy.stats import chi2, norm
import pandas as pd
import numpy as np


CONSTANTS: pd.DataFrame = pd.DataFrame.from_dict(
    dict(
        n=np.array([
            2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
            20, 21, 22, 23, 24, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75,
            80, 85, 90, 100]),
        A2=np.array([
            1.881, 1.023, 0.729, 0.577, 0.483, 0.419, 0.373, 0.337, 0.308,
            0.285, 0.266, 0.249, 0.235, 0.223, 0.212, 0.203, 0.194, 0.187,
            0.180, 0.173, 0.167, 0.162, 0.157, 0.153, 0.134, 0.120, 0.110,
            0.101, 0.094, 0.088, 0.083, 0.079, 0.075, 0.072, 0.069, 0.066,
            0.064, 0.060]),
        A3=np.array([
            2.659, 1.954, 1.628, 1.427, 1.287, 1.182, 1.099, 1.032, 0.975,
            0.927, 0.886, 0.850, 0.817, 0.789, 0.763, 0.739, 0.718, 0.698,
            0.680, 0.663, 0.647, 0.633, 0.619, 0.606]),
        B3=np.array([
            0, 0, 0, 0, 0.030, 0.118, 0.185, 0.239, 0.284, 0.321, 0.354,
            0.382, 0.406, 0.428, 0.448, 0.466, 0.482, 0.497, 0.510, 0.523,
            0.534, 0.545, 0.555, 0.565]),
        B4=np.array([
            3.267, 2.568, 2.266, 2.089, 1.970, 1.882, 1.815, 1.761, 1.716,
            1.679, 1.646, 1.618, 1.594, 1.572, 1.552, 1.534, 1.518, 1.503,
            1.490, 1.477, 1.466, 1.455, 1.445, 1.435]),
        c2=np.array([
            0.5642, 0.7236, 0.7979, 0.8407, 0.8686, 0.8882, 0.9027, 0.9139,
            0.9227, 0.9300, 0.9359, 0.9410, 0.9453, 0.9490, 0.9523, 0.9551,
            0.9576, 0.9599, 0.9619, 0.9638, 0.9655, 0.9670, 0.9684, 0.9695,
            0.9748, 0.9784, 0.9811, 0.9832, 0.9849, 0.9863, 0.9874, 0.9884,
            0.9892, 0.9900, 0.9906, 0.9911, 0.9916, 0.9925]),
        c4=np.array([
            0.7979, 0.8862, 0.9213, 0.9400, 0.9515, 0.9594, 0.9650, 0.9693,
            0.9727, 0.9754, 0.9776, 0.9794, 0.9810, 0.9823, 0.9835, 0.9845,
            0.9854, 0.9862, 0.9869, 0.9876, 0.9882, 0.9887, 0.9892, 0.9896]),
        d2=np.array([
            1.128, 1.693, 2.059, 2.326, 2.534, 2.704, 2.847, 2.970, 3.078,
            3.173, 3.258, 3.336, 3.407, 3.472, 3.532, 3.588, 3.640, 3.689,
            3.735, 3.778, 3.819, 3.858, 3.895, 3.931, 4.086, 4.213, 4.322,
            4.415, 4.498, 4.572, 4.639, 4.699, 4.755, 4.806, 4.854, 4.898,
            4.939, 5.015]),
        d3=np.array([
            0.8525, 0.8884, 0.8798, 0.8641, 0.8480, 0.8332, 0.8198, 0.8078,
            0.7971, 0.7873, 0.7785, 0.7704, 0.7630, 0.7562, 0.7499, 0.7441,
            0.7386, 0.7335, 0.7287, 0.7272, 0.7199, 0.7159, 0.7121, 0.7084,
            0.6927, 0.6799, 0.6692, 0.6601, 0.6521, 0.6452, 0.6389, 0.6333,
            0.6283, 0.6236, 0.6194, 0.6154, 0.6118, 0.6052]),
        D3=np.array([
            'NaN', 'NaN', 'NaN', 'NaN', 'NaN', 0.076, 0.136, 0.184, 0.223,
            0.256, 0.283, 0.307, 0.328, 0.347, 0.363, 0.378, 0.391, 0.403,
            0.415, 0.423, 0.434, 0.443, 0.452, 0.459, 0.491, 0.516, 0.535,
            0.551, 0.565, 0.577, 0.587, 0.596, 0.604, 0.611, 0.617, 0.623,
            0.628, 0.638]),
        d4=np.array([
            0.954, 1.588, 1.978, 2.257, 2.472, 2.645, 2.791, 2.915, 3.024,
            3.121, 3.207, 3.285, 3.356, 3.422, 3.482, 3.538, 3.591, 3.640,
            3.686, 3.730, 3.771, 3.811, 3.847, 3.883, 4.037, 4.166, 4.274,
            4.372, 4.450, 4.521, 4.591, 4.649, 4.707, 4.757, 4.806, 4.849,
            4.892, 4.968]),
        D4=np.array([
            3.267, 2.574, 2.282, 2.114, 2.004, 1.924, 1.864, 1.816, 1.777,
            1.744, 1.717, 1.693, 1.672, 1.653, 1.637, 1.622, 1.609, 1.597,
            1.585, 1.577, 1.566, 1.557, 1.548, 1.541, 1.509, 1.484, 1.465,
            1.449, 1.435, 1.423, 1.413, 1.404, 1.396, 1.389, 1.383, 1.377,
            1.372, 1.362]),
        E2=np.array([
            2.660, 1.772, 1.457, 1.290, 1.184, 1.109, 1.054, 1.010, 0.975,
            0.945, 0.921, 0.899, 0.881, 0.864, 0.849, 0.836, 0.824, 0.813,
            0.803, 0.794, 0.786, 0.778, 0.770, 0.763, 0.734, 0.712, 0.694,
            0.680, 0.667, 0.656, 0.647, 0.638, 0.631, 0.624, 0.618, 0.612,
            0.607, 0.598])
    ),
    orient='index').transpose().set_index('n')


def cp(
    average: float | int,
    std_devn: float | int,
    subgroup_size: int,
    number_subgroups: int,
    lower_spec: float | int,
    upper_spec: float | int,
    alpha: float = 0.05,
) -> tuple[float, float, float]:
    """
    Cp compares the width of the process specification to the width of the
    process variation. It does not take into consideration the deviation from
    the average. It "assumes" the process is centred between the specification
    limits. The standard deviation estimate is taken from a range or moving
    range control chart.

    Parameters
    ----------
    average : float | int,
        The average of the process.
    std_devn : float | int,
        The standard deviation of the process. It should be the "sample
        standard deviation".
    subgroup_size : int,
        This is the number of values in a control chart subgroup
    number_subgroups : int,
        This is the number of subgroups.
    lower_spec : float | int,
        The lower specification value.
    upper_spec : float | int,
        The upper specification value.
    alpha : float = 0.05
        The alpha value for the confidence interval calculations. An alpha of
        0.05 is used for a 95 % confidence interval.

    Returns
    -------
    tuple[float, float, float]
        A tuple of the capability, the lower confidence bound, and the upper confidence bound.

            - capability : float
                The Pp process capability value.
            - lower_bound : float
                The lower value of the confidence interval for Pp.
            - upper_bound : float
                The upper value of the confidence interval for Pp.

    Example
    -------
    >>> import dawgdad as dd
    >>> average = 0.11001
    >>> std_devn = 0.89312
    >>> subgroup_size = 1
    >>> number_subgroups = 39
    >>> lower_spec = -4
    >>> upper_spec = 4
    >>> alpha = 0.05
    >>> result = dd.cp(
    >>>     average=average,
    >>>     std_devn=std_devn,
    >>>     subgroup_size=subgroup_size,
    >>>     number_subgroups=number_subgroups,
    >>>     lower_spec=lower_spec,
    >>>     upper_spec=upper_spec,
    >>>     alpha=alpha
    >>> )
    (1.4928938253911381, 1.141174267641542, 1.8439148118984439)
    """
    capability = (upper_spec - lower_spec) / (6 * std_devn)
    constant_name = ["d2", "d3"]
    d2, d3 = CONSTANTS.loc[subgroup_size, constant_name]
    # as per wheeler in advanced topics of SPC
    degrees_of_freedom = (d2**2 * number_subgroups) / (2 * d3**2) + 0.2
    chi2_lower = chi2.ppf(q=(alpha / 2), df=degrees_of_freedom)
    chi2_upper = chi2.ppf(q=(1 - alpha / 2), df=degrees_of_freedom)
    lower_bound = capability * math.sqrt(chi2_lower / degrees_of_freedom)
    upper_bound = capability * math.sqrt(chi2_upper / degrees_of_freedom)
    return (capability, lower_bound, upper_bound)


def cpk(
    average: float | int,
    std_devn: float | int,
    subgroup_size: int,
    number_subgroups: int,
    lower_spec: float | int,
    upper_spec: float | int,
    alpha: float = 0.05,
    toler: float | int = 6,
) -> tuple[float, float, float, float, float]:
    """
    Cpk compares the width of the process specification to the width of the
    process variation. It takes into consideration the deviation from
    the average. The standard deviation estimate is taken from a range or
    moving range control chart.

    Parameters
    ----------
    average : float | int,
        The average of the process.
    std_devn : float | int,
        The standard deviation of the process. It should be the "sample
        standard deviation".
    subgroup_size : int,
        This is the number of values in a control chart subgroup
    number_subgroups : int,
        This is the number of subgroups.
    lower_spec : float | int,
        The lower specification value.
    upper_spec : float | int,
        The upper specification value.
    alpha : float = 0.05
        The alpha value for the confidence interval calculations. An alpha of
        0.05 is used for a 95 % confidence interval.
    toler : float | int = 6
        The multiplier of the standard deviation tolerance.

    Returns
    -------
    tuple[float, float, float, float, float]
        A tuple of the capability, the lower Cpk, the upper Cpk, the lower confidence bound, and the upper confidence bound.

            - capability : float
                The Cpk process capability value.
            - cpk_lower : float
                The Ppk value for left of the average,
            - cpk_lower : float
                The Ppk value for right of the average,
            - lower_bound : float
                The lower value of the confidence interval for Cpk.
            - upper_bound : float
                The upper value of the confidence interval for Cpk.

    Example
    -------
    >>> import dawgdad as dd
    >>> average = 0.11001
    >>> std_devn = 0.89312
    >>> subgroup_size = 2
    >>> number_subgroups = 39
    >>> lower_spec = -4
    >>> upper_spec = 4
    >>> alpha = 0.05
    >>> result = dd.cpk(
    >>>     average=average,
    >>>     std_devn=std_devn,
    >>>     subgroup_size=subgroup_size,
    >>>     number_subgroups=number_subgroups,
    >>>     lower_spec=lower_spec,
    >>>     upper_spec=upper_spec,
    >>>     alpha=alpha,
    >>>     toler=6,
    >>> )
    (
        1.4518355129583185, 1.533952137823958, 1.4518355129583185,
        1.0928917337156085, 1.8107792922010284
    )
    """
    constant_name = ["d2", "d3"]
    d2, d3 = CONSTANTS.loc[subgroup_size, constant_name]
    # as per wheeler in advanced topics of SPC
    degrees_of_freedom = (d2**2 * number_subgroups) / (2 * d3**2) + 0.2
    cpk_lower = (average - lower_spec) / (3 * std_devn)
    cpk_upper = (upper_spec - average) / (3 * std_devn)
    capability = min(cpk_lower, cpk_upper)
    z = norm.ppf(q=(1 - alpha / 2))
    lower_bound = capability - z * math.sqrt(
        (1 / (((toler / 2) ** 2) * number_subgroups))
        + ((capability**2) / (2 * degrees_of_freedom))
    )
    upper_bound = capability + z * math.sqrt(
        (1 / (((toler / 2) ** 2) * number_subgroups))
        + ((capability**2) / (2 * degrees_of_freedom))
    )
    return (capability, cpk_lower, cpk_upper, lower_bound, upper_bound)


def cpm(
    average: float | int,
    std_devn: float | int,
    sample_size: int,
    target: float | int,
    lower_spec: float | int,
    upper_spec: float | int,
    alpha: float = 0.05,
) -> tuple[float, float]:
    """
    Ppk and Cpk calculate process capability with respect to the deviation from
    the average. If a process average is not equal to the specification target,
    the process capability is not as good as one would assume. Cpm calculates
    process capability with respect to the deviation from the average and the
    the deviation from the target. The Cpm formula is closely related to the
    Taguchi Loss Function. The standard deviation uses the "population standard
    deviation."

    Parameters
    ----------
    average : float | int,
        The average of the process.
    std_devn : float | int,
        The standard deviation of the process. It should be the "sample
        standard deviation".
    sample_size : int,
        This is the sample size for the data being analysed.
    target : float
        It is the target value of the product stream.
    lower_spec : float | int,
        The lower specification value.
    upper_spec : float | int,
        The upper specification value.
    alpha : float = 0.05
        The alpha value for the confidence interval calculations. An alpha of
        0.05 is used for a 95 % confidence interval.

    Returns
    -------
    tuple[float, float]
        A tuple of the capability and the lower confidence bound.

            - capability : float
                The Cpm process capability value.
            - lower_bound : float
                The lower value of the confidence interval for Cpm.

    Example
    -------
    >>> import dawgdad as dd
    >>> average = 0.11001
    >>> std_devn = 0.868663
    >>> sample_size = 40
    >>> target = 0
    >>> lower_spec = -4
    >>> upper_spec = 4
    >>> alpha = 0.05
    >>> result = dd.cpm(
    >>>     average=average,
    >>>     std_devn=std_devn,
    >>>     sample_size= sample_size,
    >>>     target=target,
    >>>     lower_spec=lower_spec,
    >>>     upper_spec=upper_spec,
    >>>     alpha=alpha
    >>> )
    (1.5227631097133512, 1.2396924251472865)
    """
    capability = min(target - lower_spec, upper_spec - target) / (
        3 * math.sqrt(std_devn**2 + (average - target) ** 2)
    )
    aratio = (average - target) / std_devn
    degrees_of_freedom = (
        sample_size * (1 + aratio**2) ** 2 / (1 + 2 * aratio**2)
    )
    chi2_lower = chi2.ppf(q=alpha, df=degrees_of_freedom)
    lower_bound = capability * math.sqrt(chi2_lower / degrees_of_freedom)
    return (capability, lower_bound)


def pp(
    average: float | int,
    std_devn: float | int,
    sample_size: int,
    lower_spec: float | int,
    upper_spec: float | int,
    alpha: float = 0.05,
) -> tuple[float, float, float]:
    """
    Pp compares the width of the process specification to the width of the
    process variation. It does not take into consideration the deviation from
    the average. It "assumes" the process is centred between the specification
    limits. The standard deviation uses the "sample standard deviation"
    formula.

    Parameters
    ----------
    average : float | int,
        The average of the process.
    std_devn : float | int,
        The standard deviation of the process. It should be the "sample
        standard deviation".
    sample_size : int,
        This is the sample size for the data being analysed.
    lower_spec : float | int,
        The lower specification value.
    upper_spec : float | int,
        The upper specification value.
    alpha : float = 0.05
        The alpha value for the confidence interval calculations. An alpha of
        0.05 is used for a 95 % confidence interval.

    Returns
    -------
    tuple[float, float, float]
        A tuple of the capability, the lower confidence bound, and the upper confidence bound.

            - capability : float
                The Pp process capability value.
            - lower_bound : float
                The lower value of the confidence interval for Pp.
            - upper_bound : float
                The upper value of the confidence interval for Pp.

    Example
    -------
    >>> import dawgdad as dd
    >>> average = 0.11001
    >>> std_devn = 0.868663
    >>> sample_size = 40
    >>> lower_spec = -4
    >>> upper_spec = 4
    >>> alpha = 0.05
    >>> result = dd.pp(
    >>>     average=average,
    >>>     std_devn=std_devn,
    >>>     sample_size=sample_size,
    >>>     lower_spec=lower_spec,
    >>>     upper_spec=upper_spec,
    >>>     alpha=alpha
    >>> )
    (1.5349258956964131, 1.1953921108301047, 1.873778000024199)
    """
    capability = (upper_spec - lower_spec) / (6 * std_devn)
    degrees_of_freedom = sample_size - 1
    chi2_lower = chi2.ppf(q=alpha / 2, df=degrees_of_freedom)
    chi2_upper = chi2.ppf(q=1 - alpha / 2, df=degrees_of_freedom)
    lower_bound = capability * math.sqrt(chi2_lower / degrees_of_freedom)
    upper_bound = capability * math.sqrt(chi2_upper / degrees_of_freedom)
    return (capability, lower_bound, upper_bound)


def ppk(
    average: float | int,
    std_devn: float | int,
    sample_size: int,
    lower_spec: float | int,
    upper_spec: float | int,
    alpha: float = 0.05,
    toler: float | int = 6,
) -> tuple[float, float, float, float, float]:
    """
    Ppk compares the width of the process specification to the width of the
    process variation. It does take into consideration the deviation from
    the average. The standard deviation uses the "sample standard deviation"
    formula.

    Parameters
    ----------
    average : float | int,
        The average of the process.
    std_devn : float | int,
        The standard deviation of the process. It should be the "sample
        standard deviation".
    sample_size : int,
        This is the sample size for the data being analysed.
    lower_spec : float | int,
        The lower specification value.
    upper_spec : float | int,
        The upper specification value.
    alpha : float = 0.05
        The alpha value for the confidence interval calculations. An alpha of
        0.05 is used for a 95 % confidence interval.
    toler : float | int = 6
        The multiplier of the standard deviation tolerance.

    Returns
    -------
    tuple[float, float, float, float, float]
        A tuple of the capability, the lower Ppk, the upper Ppk, the lower confidence bound, and the upper confidence bound.

            - capability : float
                The Ppk process capability value.
            - ppk_lower : float
                The ppk value for left of the average,
            - ppk_lower : float
                The ppk value for right of the average,
            - lower_bound : float
                The lower value of the confidence interval for Ppk.
            - upper_bound : float
                The upper value of the confidence interval for Ppk.

    Example
    -------
    >>> import dawgdad as dd
    >>> average = 0.11001
    >>> std_devn = 0.868663
    >>> sample_size = 40
    >>> lower_spec = -4
    >>> upper_spec = 4
    >>> alpha = 0.05
    >>> result = dd.ppk(
    >>>     average=average,
    >>>     std_devn=std_devn,
    >>>     sample_size=sample_size,
    >>>     lower_spec=lower_spec,
    >>>     upper_spec=upper_spec,
    >>>     alpha=alpha,
    >>>     toler=6
    >>> (
        1.4927115962500226, 1.5771401951428037, 1.4927115962500226,
        1.1457133294762083, 1.8397098630238369
    )
    """
    degrees_of_freedom = sample_size - 1
    ppk_lower = (average - lower_spec) / (3 * std_devn)
    ppk_upper = (upper_spec - average) / (3 * std_devn)
    capability = min(ppk_lower, ppk_upper)
    z = norm.ppf(q=(1 - alpha / 2))
    lower_bound = capability - z * math.sqrt(
        (1 / (((toler / 2) ** 2) * sample_size))
        + ((capability**2) / (2 * degrees_of_freedom))
    )
    upper_bound = capability + z * math.sqrt(
        (1 / (((toler / 2) ** 2) * sample_size))
        + ((capability**2) / (2 * degrees_of_freedom))
    )
    return (capability, ppk_lower, ppk_upper, lower_bound, upper_bound)


__all__ = (
    "cp",
    "cpk",
    "cpm",
    "pp",
    "ppk",
)



"""
openpyxl functions
"""

from pathlib import Path
from typing import IO
import time
import sys
import io

from openpyxl.styles import Alignment, Border, Font, NamedStyle, PatternFill, \
    Side
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.workbook.workbook import Workbook
from dawgdad import report_summary, html_end
from openpyxl.utils import get_column_letter
from openpyxl import load_workbook
import pandas as pd
import numpy as np


def autofit_column_width(
    *,
    ws: Worksheet,
    extra_width: int
) -> Worksheet:
    """
    Autofit all columns in a worksheet.

    Parameters
    ----------
    ws : Worksheet
        The worksheet in which to autofit all columns.
    extra_width : int
        An integer to add extra width so that the column edges are not flush.

    Returns
    -------
    ws : Worksheet
        The worksheet in which autofit was applied to all columns

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.autofit_column_width(
    ...     ws=ws,
    ...     extra_width=7
    ... )
    """
    for column in ws.columns:
        max_width = 0
        column_letter = get_column_letter(column[0].column)
        for cell in column:
            try:  # Necessary to avoid error on empty cells
                if len(str(cell.value)) > max_width:
                    max_width = len(cell.value)
            except Exception:
                pass
        adjusted_width = (max_width + extra_width)
        ws.column_dimensions[column_letter].width = adjusted_width
    return ws


def cell_fill_down(
    *,
    ws: Worksheet,
    min_row: int,
    max_row: int,
    min_col: int,
    max_col: int
) -> Worksheet:
    """
    Fill empty cell with the value from the cell above

    Parameters
    ----------
    ws : Worksheet
        The worksheet in which to change the case of column(s).
    min_row : int
        The first row in the range to change.
    max_row : int
        The last row in the range to change.
    min_col : int
        The first column in the range to change.
    max_col : int
        The last column in the range to change.

    Returns
    -------
    ws : Worksheet
        The worksheet in which cells were modified.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> fill_down_columns = ["col1", "col2", "col3"]
    >>> for column in fill_down_columns:
    ...     ws = dd.cell_fill_down(
    ...         ws=ws,
    ...         min_row=2,
    ...         max_row=ws.max_row,
    ...         min_col=1,
    ...         max_col=3
    ...     )
    """
    row_count = 0
    for row in ws.iter_rows(
        min_col=min_col,
        max_col=max_col
    ):
        for cell in row:
            if cell.value:
                row_count += 1
    if row_count > 0:
        for row in ws.iter_rows(
            min_row=min_row + 1,  # start one row below the 'start' row
            max_row=max_row,
            min_col=min_col,
            max_col=max_col
        ):
            for cell in row:
                if cell.value in [None, 'None', '']:
                    cell.value = ws[cell.row - 1][min_col - 1].value
    return ws


def cell_style(
    *,
    wb: Workbook,
    style_name: str = 'cell_style',
    font_name: str = 'Lucida Sans',
    font_size: int = 11,
    font_bold: bool = True,
    font_colour: str = '000000',
    horizontal_alignment: str = 'center',
    vertical_alignment: str = 'center',
    wrap_text: str | bool = None,
    fill_type: str | bool = 'solid',
    foreground_colour: str | bool = 'd9d9d9',
    border_style: str | bool = None,
    border_colour: str | bool = None,
    number_format: str | bool = None
) -> NamedStyle:
    """
    Define a cell style

    Parameters
    ----------
    wb : Workbook
        The workbook in which to define the cell style.
    style_name : str = 'cell_style'
        The name for the cell style.
    font_name : str = 'Lucida Sans'
        The font name for the style.
    font_size : int = 11
        The font size for the style.
    font_bold : bool = True
        A boolean or string to apply bold style.
    font_colour : str = 'ffffff'
        The string for the font colour.
    horizontal_alignment : str = 'center'
        The string for horizontal alignment.
    vertical_alignment : str = 'center'
        The string for vertical alignment.
    wrap_text : str | bool = None
        A boolean or string to wrap text.
    fill_type : str = 'solid'
        The string for the fill type.
    foreground_colour : str = 'd9d9d9'
        The string for the foreground colour.
    border_style : str | bool = None
        A boolean or string to apply a border.
    border_colour : str | bool = None
        A boolean or string to apply a border colour.
    number_format : str | bool = None
        A boolean or string to apply a number format.

    Returns
    -------
    row_style : NamedStyle
        The named style.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> red_cell_style = dd.cell_style(
    ...     wb=wb,
    ...     style_name='red_cell_style',
    ...     font_colour='ffffff',
    ...     foreground_colour='c00000'
    ... )
    >>> wb.add_named_style(red_cell_style) # doctest: +SKIP
    >>> for cell in ['C1', 'D1', 'E1']:
    ...     ws[cell].style = red_cell_style # doctest: +SKIP
    """
    cell_style = NamedStyle(name=style_name)
    cell_style.font = Font(
        name=font_name,
        size=font_size,
        bold=font_bold,
        color=font_colour
    )
    cell_style.alignment = Alignment(
        horizontal=horizontal_alignment,
        vertical=vertical_alignment,
        wrap_text=wrap_text
    )
    cell_style.fill = PatternFill(
        fill_type=fill_type,
        fgColor=foreground_colour
    )
    cell_style.border = Border(
        bottom=Side(
            border_style=border_style,
            color=border_colour
        )
    )
    cell_style.number_format = number_format
    wb.add_named_style(cell_style)
    return (wb, cell_style)


def change_case_worksheet_columns(
    *,
    ws: Worksheet,
    min_col: int,
    max_col: int,
    min_row: int,
    max_row: int,
    case: str = 'upper'
) -> Worksheet:
    """
    Change case for one or more worksheet columns.

    Parameters
    ----------
    ws : Worksheet
        The worksheet in which to change the case of column(s).
    min_col : int
        The first column in the range to change.
    max_col : int
        The last column in the range to change.
    min_row : int
        The first row in the range to change.
    max_row : int
        The last row in the range to change.
    case: str = 'upper'
        The case to change. Currently only upper or lower.

    Returns
    -------
    ws : Worksheet
        A worksheet from a workbook.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.change_case_worksheet_columns(
    ...     ws=ws,
    ...     min_col=4,
    ...     max_col=6,
    ...     min_row=1,
    ...     max_row=ws.max_row,
    ...     case='upper'
    ... )
    """
    for col in ws.iter_cols(
        min_col=min_col,
        max_col=max_col,
        min_row=min_row,
        max_row=max_row
    ):
        for cell in col:
            if case == 'upper':
                cell.value = str(cell.value).upper()
            elif case == 'lower':
                cell.value = str(cell.value).lower()
    return ws


def exit_script(
    *,
    original_stdout: IO[str],
    output_url: str
) -> None:
    """
    Exit from a script and complete the html file.

    Parameters
    ----------
    original_stdout : IO[str]
        The original stdout.
    output_url : str
        The output url.

    Example
    -------
    >>> import dawgdad as dd
    >>> output_url = '../tests/my_html_file.html'
    >>> original_stdout = dd.html_begin(output_url=output_url)
    >>> dd.exit_script(
    ...     original_stdout=original_stdout,
    ...     output_url=output_url
    ... ) # doctest: +SKIP
    """
    html_end(
        original_stdout=original_stdout,
        output_url=output_url
    )
    sys.exit()


def list_duplicate_worksheet_rows(
    *,
    ws: Worksheet,
    columns_to_ignore: list[int] = None
) -> list[int]:
    """
    Find duplicate rows in a worksheet.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    columns_to_ignore : list[int] = None
        A list of column numbers to not use in determining duplicate rows.

    Returns
    -------
    duplicate_rows : list[int]
        A list of duplicate row numbers.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> duplicate_rows = dd.list_duplicate_worksheet_rows(ws=ws)
    >>> ws = dd.remove_worksheet_rows(
    ...     ws=ws,
    ...     rows_to_remove=duplicate_rows
    ... )
    """
    duplicate_rows = []
    unique_rows = []
    for row in ws.rows:
        working_list = []
        for cell in row:
            if cell.column not in columns_to_ignore:
                working_list.append(cell.value)
        if working_list not in unique_rows:
            unique_rows.append(working_list)
        else:
            duplicate_rows.append(cell.row)
    return duplicate_rows


def list_empty_and_nan_worksheet_rows(
    *,
    ws: Worksheet,
    min_row: int
) -> list[int]:
    """
    Create list of row numbers of blank worksheet rows.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    min_row : int
        Start row for iteration.

    Returns
    -------
    blank_rows : list[int]
        List of row numbers.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> sheetname = "sheetname"
    >>> ws = wb[sheetname] # doctest: +SKIP
    >>> blank_rows = dd.list_nan_worksheet_rows(
    ...     ws=ws,
    ...     min_row=2
    ... ) # doctest: +SKIP
    """
    blank_rows = []
    for row in ws.iter_rows(min_row=min_row):
        onerow = [cell.value for cell in row]
        if all(item in [
            None, 'None', 'NONE', '', np.nan
        ] for item in onerow):
            blank_rows.append(row[0].row)
    return blank_rows


def list_empty_except_nan_worksheet_rows(
    *,
    ws: Worksheet,
    min_row: int
) -> list[int]:
    """
    Create list of row numbers of empty worksheet rows, except those
    with np.nan.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    min_row : int
        Start row for iteration.

    Returns
    -------
    empty_rows : list[int]
        List of row numbers.

    Example
    -------
    Remove empty rows starting from row 2.
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> empty_rows = dd.list_empty_except_nan_worksheet_rows(
    ...     ws=ws,
    ...     min_row=2
    ... )
    """
    empty_rows = []
    for row in ws.iter_rows(min_row=min_row):
        onerow = [cell.value for cell in row]
        if all(item == onerow[0] for item in onerow):
            empty_rows.append(row[0].row)
    return empty_rows


def list_nan_worksheet_rows(
    *,
    ws: Worksheet,
    min_row: int
) -> list[int]:
    """
    Create list of row numbers of blank worksheet rows.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    min_row : int
        Start row for iteration.

    Returns
    -------
    blank_rows : list[int]
        List of row numbers.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> sheetname = "sheetname"
    >>> ws = wb[sheetname] # doctest: +SKIP
    >>> blank_rows = dd.list_nan_worksheet_rows(
    ...     ws=ws,
    ...     min_row=2
    ... )
    """
    blank_rows = []
    for row in ws.iter_rows(min_row=min_row):
        onerow = [cell.value for cell in row]
        if all(item != item for item in onerow):
            blank_rows.append(row[0].row)
    return blank_rows


def list_rows_with_content(
    *,
    ws: Worksheet,
    min_row: int,
    column: int,
    text: str
) -> list[int]:
    """
    List rows that contain specific text in a specified column.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    min_row : int
        Start row for iteration.
    column : int
        The column to search.
    text : str
        The text to search.

    Returns
    -------
    list[int]
        A list of row numbers.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> rows_with_text = dd.list_rows_with_content(
    ...     ws=ws,
    ...     min_row=2,
    ...     column=11,
    ...     text='ETA'
    ... )
    """
    rows_with_text = []
    for row in ws.iter_rows(
        min_row=min_row,
        min_col=column,
        max_col=column
    ):
        for cell in row:
            if cell.value == text:
                rows_with_text.append(row[0].row)
    return rows_with_text


def read_workbook(
    *,
    filename: Path | str,
    data_only: bool = True
) -> tuple[Workbook, list[str]]:
    """
    Read a workbook, print the Path, and print the sheet names.

    Parameters
    ----------
    filename : Path | str
        The file containing the workbook.
    data_only : bool = True
        If True, read values stored in the cells. If False, read formulae
        stored in the cells.

    Returns
    -------
    tuple[Workbook, list[str]]
        A tuple of a Workbook and a list of sheet names.

            - wb : Workbook
                A workbook.
            - sheet_names : list[str]
                The sheet names in the workbook.

    Examples
    --------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> filename = "../tests/excel_file.xlsx"
    >>> wb, sheet_names = dd.read_workbook(
    ...     filename=filename,
    ...     data_only=True
    ... ) # doctest: +SKIP
    """
    wb = load_workbook(
        filename=filename,
        data_only=data_only
    )
    sheet_names = wb.sheetnames
    return (wb, sheet_names)


def remove_empty_worksheet_rows(
    *,
    ws: Worksheet,
    empty_rows: list[int]
) -> Worksheet:
    """
    Delete empty worksheet rows.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    empty_rows : list[int]
        List of row numbers.

    Returns
    -------
    ws : Worksheet
        A worksheet from a workbook.

    Example
    -------
    Remove empty rows found.
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.remove_empty_worksheet_rows(
    ...     ws=ws,
    ...     empty_rows=[5, 6, 7]
    ... )
    """
    for row_idx in reversed(empty_rows):
        ws.delete_rows(
            idx=row_idx,
            amount=1
        )
    return ws


def remove_worksheet_columns(
    *,
    ws: Worksheet,
    starting_column: int,
    number_of_columns: int
) -> Worksheet:
    """
    Remove worksheet columns.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    starting_column : int
        The first column to remove.
    number_of_columns : int
        The number of columns to remove.

    Returns
    -------
    ws : Worksheet
        A worksheet from a workbook.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.remove_worksheet_columns(
    ...     ws=ws,
    ...     starting_column=13,
    ...     number_of_columns=3
    ... )
    """
    ws.delete_cols(
        idx=starting_column,
        amount=number_of_columns
    )
    return ws


def remove_worksheet_rows(
    *,
    ws: Worksheet,
    rows_to_remove: list[int]
) -> Worksheet:
    """
    Remove worksheet rows.

    Parameters
    ----------
    ws : Worksheet
        A worksheet from a workbook.
    rows_to_remove: list[int]
        The list of row numbers to remove.

    Returns
    -------
    ws : Worksheet
        A worksheet from a workbook.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.remove_worksheet_rows(
    ...     ws=ws,
    ...     rows_to_remove=[4, 5, 6]
    ... )
    """
    for row in reversed(rows_to_remove):
        ws.delete_rows(
            idx=row,
            amount=1
        )
    return ws


def replace_text(
    *,
    ws: Worksheet,
    column: int,
    text: tuple[tuple[str, str]]
) -> Worksheet:
    """
    Search and replace text in a cell.

    Parameters
    ----------
    ws : Worksheet
        The worksheet in which to search and replace text.
    column : int
        The column number in which to search and replace text.
    text ; tuple[tuple[str, str]]
        The search and replace text.

    Returns
    -------
    ws : Worksheet
        The worksheet in which text was replaced.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.replace_text(
    ...     ws=ws,
    ...     column=13,
    ...     text=['old_text', 'new_text']
    ... ) # doctest: +SKIP
    """
    for row in ws.iter_rows(
        min_col=column,
        max_col=column
    ):
        for item in text:
            for cell in row:
                cell.value = str(cell.value).replace(*item)
    return ws


def validate_column_labels(
    *,
    ws: Worksheet,
    column_labels: list[str],
    first_column: int,
    last_column: int,
    row_of_labels: int,
    start_time: float = None,
    stop_time: float = None,
    original_stdout: IO[str] = None,
    output_url: str = None

) -> Worksheet:
    """
    Validate the labels of a worksheet with a desired list of labels

    Parameters
    ----------
    ws : Worksheet
        The worksheet to analyze.
    column_labels : list[str]
        The list of desired column labels.
    first_column : int
        The first column of the label range in the worksheet.
    last_column : int
        The last column of the label range in the worksheet.
    row_of_labels : int
        The row number of the labels in the worksheet.
    start_time : float = None
        The start time of the script.
    stop_time : float = None
        The stop time of the script.
    original_stdout : IO[str] = None
        The original stdout.
    output_url : str = None
        The output url.

    Returns
    -------
    ws : Worksheet,
        The worksheet to analyze.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> ws = dd.validate_column_labels(
    ...     ws=ws,
    ...     column_labels=column_labels,
    ...     first_column=first_column,
    ...     last_column=last_column,
    ...     row_of_labels=row_of_labels,
    ...     start_time=start_time,
    ...     stop_time=time.time(),
    ...     original_stdout=original_stdout,
    ...     output_url=output_url
    ... ) # doctest: +SKIP
    """
    labels_found = []
    for col in range(first_column, last_column + 1):
        labels_found.append(ws.cell(row=row_of_labels, column=col).value)
    if len(labels_found) == len(column_labels) and \
            labels_found[-1] == column_labels[-1]:
        for col, label in zip(
            range(first_column, last_column + 1),
            column_labels
        ):
            ws.cell(
                row=row_of_labels,
                column=col
            ).value = label
    elif stop_time:
        print(
            'Column labels or number of columns are incorrect. '
            'Fix. Re-run script.'
        )
        print('Labels found:')
        print(labels_found)
        print('Length of labels found:', len(labels_found))
        print('Labels should be:')
        print(column_labels)
        print('Length of labels should be:', len(column_labels))
        print('XXX File NOT OK XXX')
        stop_time = stop_time
        report_summary(
            start_time=start_time,
            stop_time=stop_time
        )
        exit_script(
            original_stdout=original_stdout,
            output_url=output_url
        )
    return ws


def unique_list_items(
    *,
    ws: Worksheet,
    row_of_labels: int,
    row_below_labels: int,
    column_name_varname: str,
    text_to_replace: list[str],
    text_to_remove: list[str]
) -> tuple[list[int], list[int]]:
    """
    Determine list of unique items in varname.
    Replace text.

    TODO:
    This function does too many things. Break it up.
    Add detail to docstring.

    Parameters
    ----------
    ws : Worksheet
        The worksheet to analyze.
    row_of_labels : int
        The row number of the labels in the worksheet.
    row_below_labels : int
        The row number below the label row.
    column_name_varname : str
        The label of a column.
    text_to_replace : list[str]
        A list of the text to search for.
    text_to_remove : list[str]
        A list of the text to remove.

    Returns
    -------
    tuple[list[int], list[int]]
        A tuple of a list of unique items and a list of column numbers.

            - varname : tuple[list[int]
                The list of unique items.
            - column_numbers : list[int]]
                The list of column numbers.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> list_part_numbers, list_column_numbers = dd.unique_list_items(
    ...     ws=ws,
    ...     row_of_labels=row_of_labels,
    ...     row_below_labels=row_below_labels,
    ...     column_name_varname=column_name,
    ...     text_to_replace=text_to_replace,
    ...     text_to_remove=text_to_remove
    ... ) # doctest: +SKIP
    """
    column_numbers = [
        col for col in range(ws.min_column, ws.max_column + 1)
    ]
    column_labels = []
    for row in ws.iter_rows(
        min_row=row_of_labels,
        max_row=row_of_labels,
        min_col=column_numbers[0],
        max_col=column_numbers[-1]
    ):
        for cell in row:
            column_labels.append(cell.value)
    column_names_numbers = dict(zip(column_labels, column_numbers))
    varname_replace = []
    for row in ws.iter_rows(
        min_row=row_below_labels,
        min_col=column_names_numbers[column_name_varname],
        max_col=column_names_numbers[column_name_varname]
    ):
        for cell in row:
            if cell.value:
                varname_replace.append(cell.value)
    varname_remove = []
    for item in varname_replace:
        for thing in text_to_replace:
            item = str(item).replace(thing, '')
        varname_remove.append(item)
    varname = [x for x in varname_remove if x not in text_to_remove]
    varname = (list(set(varname)))
    return (varname, column_numbers)


def validate_sheet_names(
    *,
    wb: Workbook,
    filename: Path | str,
    sheet_name: str,
    sheet_names: list[str],
    start_time: float,
    original_stdout: io.TextIOWrapper,
    output_url: str
) -> Workbook:
    """
    Parameters
    ----------
    wb : Workbook
        A workbook.
    filename : Path | str
        The file containing the workbook.
    sheet_name : str
        A sheet name in the workbook.
    sheet_names : list[str]
        The sheet names in the workbook.
    start_time : float
        The start time of the script.
    original_stdout : io.TextIOWrapper
        The buffered text stream for the html output.
    output_url : str
        The html filename.

    Returns
    -------
    wb : Workbook
        The workbook with a revised sheetname.

    Example
    -------
    >>> import dawgdad as dd
    >>> start_time = time.perf_counter()
    >>> stop_time = time.perf_counter()
    >>> filename = "../tests/excel_file.xlsx"
    >>> sheet_name = "sheet_01"
    >>> sheet_names = ["sheet_01", "sheet_02"]
    >>> output_url = "../tests/my_html_file.html"
    >>> original_stdout = dd.html_begin(output_url=output_url)
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> wb = validate_sheet_names(
    ...     wb=wb,
    ...     filename=filename,
    ...     sheet_name=sheet_name,
    ...     sheet_names=sheet_names,
    ...     start_time=start_time,
    ...     original_stdout=original_stdout,
    ...     output_url=output_url
    ... ) # doctest: +SKIP
    """
    if sheet_name not in sheet_names and len(sheet_names) != 1:
        print('Manually rename one of these sheets:')
        print(wb.sheetnames)
        print('Then re-run script')
        print('XXX File NOT OK XXX')
        stop_time = time.time()
        report_summary(
            start_time=start_time,
            stop_time=stop_time
        )
        exit_script(
            original_stdout=original_stdout,
            output_url=output_url
        )
    elif sheet_name not in sheet_names and len(sheet_names) == 1:
        print('One sheet found and it was re-named.')
        ws = wb.active
        ws.title = sheet_name
        wb.save(filename=filename)
    elif sheet_name in sheet_names and len(sheet_names) != 1:
        sheet_names_removed = [x for x in sheet_names if x != sheet_name]
        for sheet in sheet_names_removed:
            wb.remove(worksheet=wb[sheet])
        print('Sheet names removed:')
        print(sheet_names_removed)
        wb.save(filename=filename)
    return wb


def write_dataframe_to_worksheet(
    *,
    ws: Worksheet,
    df: pd.DataFrame,
    index: bool = False,
    header: bool = True
) -> Worksheet:
    """
    Write a dataframe to a worksheet.

    Parameters
    ----------
    ws : Worksheet
        The worksheet to which the dataframe will be written.
    df : pd.DataFrame
        The dataframe of data.
    index : bool = False
        Boolean to determine if dataframe index is written to worksheet.
    header : bool = True
        Boolean to determine if dataframe header is written to worksheet.

    Returns
    -------
    ws : Worksheet
        The worksheet created.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> df = pd.DataFrame(data={
    ...     'X': [25.0, 24.0, 35.5, np.nan, 23.1],
    ...     'Y': [27, 24, np.nan, 23, np.nan],
    ...     'Z': ['a', 'b', np.nan, 'd', 'e']
    ... })
    >>> ws = dd.write_dataframe_to_worksheet(
    ...     ws=ws,
    ...     df=df,
    ...     index=False,
    ...     header=True
    ... )
    """
    for row in dataframe_to_rows(
        df=df,
        index=index,
        header=header
    ):
        ws.append(row)
    return ws


def number_non_empty_rows(
    *,
    ws: Worksheet,
    column_number: int,
    start_row: int,
) -> int:
    """
    Determine the number of non-empty rows for a single column.

    Parameters
    ----------
    ws : Worksheet
        The worksheet to analyze.
    column_number : int
        The desired column number.
    start_row : int
        The row at which to start evaluating cells.

    Returns
    -------
    row_count : int
        The number of non-empty rows.

    Example
    -------
    >>> import dawgdad as dd
    >>> wb = Workbook()
    >>> ws = wb.active
    >>> start_row = 2
    >>> column_number = 1
    >>> row_count = dd.number_non_empty_rows(
    ...     ws=ws,
    ...     column_number=column_number,
    ...     start_row=start_row,
    ... )
    """
    row_count = 0
    for row in ws.iter_rows(
        min_row=start_row,
        min_col=column_number,
        max_col=column_number
    ):
        for cell in row:
            if cell.value:
                row_count += 1
    return row_count


__all__ = (
    'list_empty_except_nan_worksheet_rows',
    'list_empty_and_nan_worksheet_rows',
    'list_duplicate_worksheet_rows',
    'change_case_worksheet_columns',
    'write_dataframe_to_worksheet',
    'remove_empty_worksheet_rows',
    'remove_worksheet_columns',
    'list_nan_worksheet_rows',
    'list_rows_with_content',
    'validate_column_labels',
    'number_non_empty_rows',
    'remove_worksheet_rows',
    'validate_sheet_names',
    'autofit_column_width',
    'unique_list_items',
    'cell_fill_down',
    'read_workbook',
    'replace_text',
    'exit_script',
    'cell_style',
)



#! /usr/bin/env python3
"""
Example of box-and-whisker plot.
"""

from pathlib import Path

import dawgdad as dd


def main():
    dd.style_graph()
    series = dd.random_data()
    fig, ax = dd.plot_boxplot(series=series)
    ax.set_title(label='Box-and-whisker plot')
    ax.set_xticks(ticks=[1], labels=['series'])
    ax.set_ylabel('y')
    fig.savefig(fname=Path('boxplot.svg'))


if __name__ == '__main__':
    main()



import warnings

from pathlib import Path, PosixPath

import dawgdad as dd
import pandas as pd
import numpy as np


pd.set_option('future.no_silent_downcasting', True)
df = pd.DataFrame(
    data=dict(
        floats=[1.0, np.NaN, 3.0, np.NaN, 5.0, 6.0, np.NaN],
        text=["A", "B", "C", "D", "E", "F", np.NaN],
        dates=[
            "1956-06-08", "1956-06-08",
            "1956-06-08", "1956-06-08",
            "1956-06-08", "1956-06-08",
            pd.NaT
        ],
        all_nan=[np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
        all_nat=[pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT],
        all_none=[None, None, None, None, None, None, None],
        all_space=["", " ", "", " ", "", "", ""],
        nan_space=[np.NaN, "", " ", np.NaN, np.NaN, np.NaN, np.NaN],
        nan_none=[np.NaN, None, np.NaN, np.NaN, None, np.NaN, None],
        mixed=[None, np.NaN, pd.NaT, pd.NaT, None, np.NaN, pd.NaT],
        integers=[1, 2, np.NaN, 4, 5, 6, np.NaN],
    )
).replace(
    r"^\s+$",
    np.NaN,
    regex=True
).replace(
    "",
    np.NaN,
    regex=True
).astype(
    dtype={
        "integers": "Int64",
        "floats": "float64",
        "text": "object",
        "dates": "datetime64[ns]",
        "all_nan": "float64",
        "all_nat": "datetime64[ns]",
        "all_none": "float64",
        "all_space": "float64",
        "nan_space": "float64",
        "nan_none": "float64",
        "mixed": "datetime64[ns]"
    }
)


def test_listone_contains_all_listtwo_substrings():
    pass


def test_number_empty_cells_in_columns():
    pass


def test_convert_seconds_to_hh_mm_ss():
    """
    Convert seconds.to hours, minutes and seconds.
    """
    result = dd.convert_seconds_to_hh_mm_ss(seconds=1)
    expected = (0, 0, 1)
    assert result == expected
    result = dd.convert_seconds_to_hh_mm_ss(seconds=61)
    expected = (0, 1, 1)
    assert result == expected
    result = dd.convert_seconds_to_hh_mm_ss(seconds=3601)
    expected = (1, 0, 1)
    assert result == expected
    result = dd.convert_seconds_to_hh_mm_ss(seconds=3661)
    expected = (1, 1, 1)
    result = dd.convert_seconds_to_hh_mm_ss(seconds=251)
    expected = (0, 4, 11)
    assert result == expected


def test_parameters_dict_replacement():
    pass


def test_parameters_text_replacement():
    pass


def test_ask_save_as_file_name_path():
    pass


def test_optimize_datetime_columns():
    pass


def test_optimize_integer_columns():
    pass


def test_print_dictionary_by_key():
    pass


def test_optimize_object_columns():
    pass


def test_ask_open_file_name_path():
    pass


def test_convert_csv_to_feather():
    pass


def test_find_int_float_columns():
    pass


def test_find_timedelta_columns():
    pass


def test_optimize_float_columns():
    pass


def test_create_dataframe_norm():
    pass


def test_replace_column_values():
    pass


def test_feature_percent_empty():
    pass


def test_find_category_columns():
    pass


def test_find_datetime_columns():
    pass


def test_list_one_list_two_ops():
    """
    Create a list of items comparing two lists:
    - Items unique to list_one
    - Items unique to list_two
    - Items common to both lists (intersection)
    Duplicate items are not removed.
    """
    list_one = [1, 2, 3, 4, 5, 6]
    list_two = [4, 5, 6, 7, 8, 9]
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_one"
    )
    expected = [1, 2, 3]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_two"
    )
    expected = [7, 8, 9]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="intersection"
    )
    expected = [4, 5, 6]
    assert result == expected
    list_one = ["mo", "larry", "curly", "curly-joe", "shemp"]
    list_two = ["curly-joe", "shemp", "tom", "dick", "harry"]
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_one"
    )
    expected = ["mo", "larry", "curly"]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_two"
    )
    expected = ["tom", "dick", "harry"]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="intersection"
    )
    expected = ["curly-joe", "shemp"]
    assert result == expected
    list_one = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    list_two = [4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_one"
    )
    expected = [1.0, 2.0, 3.0]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_two"
    )
    expected = [7.0, 8.0, 9.0]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="intersection"
    )
    expected = [4.0, 5.0, 6.0]
    assert result == expected
    list_one = [1, 2, 3.0, 4.0, 5, "mo", "larry", "curly-joe"]
    list_two = [2, 3, 4.0, 5.0, 6.0, "mo", "larry", "shemp"]
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_one"
    )
    expected = [1, 5, 3.0, "curly-joe"]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="list_two"
    )
    expected = [3, 5.0, 6.0, "shemp"]
    assert result == expected
    result = dd.list_one_list_two_ops(
        list_one=list_one,
        list_two=list_two,
        action="intersection"
    )
    expected = [2, 4.0, "mo", "larry"]
    assert result == expected


def test_series_replace_string():
    pass


def test_delete_empty_columns():
    """
    Test that all elements of a column:
    - are empty for all columns
    - are empty for specific columns
    """
    result1 = dd.delete_empty_columns(df=df)
    expected1 = pd.DataFrame(
        data=dict(
            floats=[1.0, np.NaN, 3.0, np.NaN, 5.0, 6.0, np.NaN],
            text=["A", "B", "C", "D", "E", "F", np.NaN],
            dates=[
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                pd.NaT
            ],
            integers=[1, 2, np.NaN, 4, 5, 6, np.NaN],
        )
    ).replace(
        r"^\s+$",
        np.NaN,
        regex=True
    ).replace(
        "",
        np.NaN,
        regex=True
    ).astype(
        dtype={
            "integers": "Int64",
            "floats": "float64",
            "text": "object",
            "dates": "datetime64[ns]",
        }
    )
    assert result1.equals(other=expected1)
    list_empty_columns = ["mixed", "nan_none"]
    # Delete columns using list_empty_columns
    result2 = dd.delete_empty_columns(
        df=df,
        list_empty_columns=list_empty_columns
    )
    expected2 = pd.DataFrame(
        data=dict(
            floats=[1.0, np.NaN, 3.0, np.NaN, 5.0, 6.0, np.NaN],
            text=["A", "B", "C", "D", "E", "F", np.NaN],
            dates=[
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                pd.NaT
            ],
            all_nan=[np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
            all_nat=[pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT],
            all_none=[None, None, None, None, None, None, None],
            all_space=["", " ", "", " ", "", "", ""],
            nan_space=[np.NaN, "", " ", np.NaN, np.NaN, np.NaN, np.NaN],
            integers=[1, 2, np.NaN, 4, 5, 6, np.NaN],
        )
    ).replace(
        r"^\s+$",
        np.NaN,
        regex=True
    ).replace(
        "",
        np.NaN,
        regex=True
    ).astype(
        dtype={
            "integers": "Int64",
            "floats": "float64",
            "text": "object",
            "dates": "datetime64[ns]",
            "all_nan": "float64",
            "all_nat": "datetime64[ns]",
            "all_none": "float64",
            "all_space": "float64",
            "nan_space": "float64"
        }
    )
    assert result2.equals(other=expected2)
    # No not delete columns using list_empty_columns because
    # not all columns in list are empty
    list_empty_columns = ["mixed", "nan_none", "integers"]
    result3 = dd.delete_empty_columns(
        df=df,
        list_empty_columns=list_empty_columns
    )
    expected3 = pd.DataFrame(
        data=dict(
            floats=[1.0, np.NaN, 3.0, np.NaN, 5.0, 6.0, np.NaN],
            text=["A", "B", "C", "D", "E", "F", np.NaN],
            dates=[
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                pd.NaT
            ],
            all_nan=[np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
            all_nat=[pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT],
            all_none=[None, None, None, None, None, None, None],
            all_space=["", " ", "", " ", "", "", ""],
            nan_space=[np.NaN, "", " ", np.NaN, np.NaN, np.NaN, np.NaN],
            nan_none=[np.NaN, None, np.NaN, np.NaN, None, np.NaN, None],
            mixed=[None, np.NaN, pd.NaT, pd.NaT, None, np.NaN, pd.NaT],
            integers=[1, 2, np.NaN, 4, 5, 6, np.NaN],
        )
    ).replace(
        r"^\s+$",
        np.NaN,
        regex=True
    ).replace(
        "",
        np.NaN,
        regex=True
    ).astype(
        dtype={
            "integers": "Int64",
            "floats": "float64",
            "text": "object",
            "dates": "datetime64[ns]",
            "all_nan": "float64",
            "all_nat": "datetime64[ns]",
            "all_none": "float64",
            "all_space": "float64",
            "nan_space": "float64",
            "nan_none": "float64",
            "mixed": "datetime64[ns]"
        }
    )
    assert result3.equals(other=expected3)


def test_directory_file_print():
    pass


def test_replace_text_numbers():
    pass


def test_find_integer_columns():
    pass


def test_find_object_columns():
    pass


def test_rename_some_columns():
    pass


def test_series_memory_usage():
    pass


def test_ask_directory_path():
    pass


def test_rename_all_columns():
    pass


def test_find_float_columns():
    pass


def test_remove_punctuation():
    pass


def test_print_list_by_item():
    pass


def test_delete_empty_rows():
    """
    Test delete empty rows:
    - all elements for a row for all columns
    - all elements for a row for specific columns
    """
    # Delete columns where all elements of a column are empty
    result = dd.delete_empty_rows(df=df)
    expected = pd.DataFrame(
        data=dict(
            floats=[1.0, np.NaN, 3.0, np.NaN, 5.0, 6.0],
            text=["A", "B", "C", "D", "E", "F"],
            dates=[
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08",
                "1956-06-08", "1956-06-08"
            ],
            all_nan=[np.NaN, np.NaN, np.NaN, np.NaN, np.NaN, np.NaN],
            all_nat=[pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT, pd.NaT],
            all_none=[None, None, None, None, None, None],
            all_space=["", " ", "", " ", "", ""],
            nan_space=[np.NaN, "", " ", np.NaN, np.NaN, np.NaN],
            nan_none=[np.NaN, None, np.NaN, np.NaN, None, np.NaN],
            mixed=[None, np.NaN, pd.NaT, pd.NaT, None, np.NaN],
            integers=[1, 2, np.NaN, 4, 5, 6],
        )
    ).replace(
        r"^\s+$",
        np.NaN,
        regex=True
    ).replace(
        "",
        np.NaN,
        regex=True
    ).astype(
        dtype={
            "integers": "Int64",
            "floats": "float64",
            "text": "object",
            "dates": "datetime64[ns]",
            "all_nan": "float64",
            "all_nat": "datetime64[ns]",
            "all_none": "float64",
            "all_space": "float64",
            "nan_space": "float64",
            "nan_none": "float64",
            "mixed": "datetime64[ns]"
        }
    )
    assert result.equals(other=expected)


def test_delete_list_files():
    pattern_startswith = ["job_aids"]
    path = "dir_directories"
    result1 = dd.list_directories(path=path)
    expected1 = [
        "cheatsheet_directory", "another_directory", "job_aids_directory"
    ]
    assert set(result1) == set(expected1)
    result2 = dd.list_directories(
        path=path,
        pattern_startswith=pattern_startswith
    )
    expected2 = ["job_aids_directory"]
    assert set(result2) == set(expected2)
    pattern_startswith = ["job_aids", "cheatsheet"]
    result3 = dd.list_directories(
        path=path,
        pattern_startswith=pattern_startswith
    )
    expected3 = ["cheatsheet_directory", "job_aids_directory"]
    assert set(result3) == set(expected3)


def test_find_bool_columns():
    pass


def test_create_dataframe():
    pass


def test_create_directory():
    pass


def test_delete_directory():
    pass


def test_list_change_case():
    pass


def test_list_directories():
    pattern_startswith = ["job_aids"]
    path = "dir_directories"
    result1 = dd.list_directories(path=path)
    expected1 = [
        'cheatsheet_directory', 'another_directory', 'job_aids_directory'
    ]
    assert set(result1) == set(expected1)
    result2 = dd.list_directories(
        path=path,
        pattern_startswith=pattern_startswith
    )
    expected2 = ['job_aids_directory']
    assert set(result2) == set(expected2)
    pattern_startswith = ["job_aids", "cheatsheet"]
    result3 = dd.list_directories(
        path=path,
        pattern_startswith=pattern_startswith
    )
    expected3 = ['cheatsheet_directory', 'job_aids_directory']
    assert set(result3) == set(expected3)


def test_optimize_columns():
    pass


def test_rename_directory():
    pass


def test_process_columns():
    pass


def test_copy_directory():
    pass


def test_dataframe_info():
    pass


def test_delete_columns():
    pass


def test_quit_sap_excel():
    pass


def test_mask_outliers():
    pass


def test_process_rows():
    pass


def test_delete_rows():
    pass


def test_list_files():
    pattern_extension = [".html", ".HTML", ".mkd", ".MKD"]
    pattern_startswith = ["file_", "job_aid_"]
    directory = "dir_files"
    result1 = sorted(dd.list_files(directory=directory))
    expected1 = sorted([
        PosixPath("dir_files/job_aid_file_one.html"),
        PosixPath("dir_files/job_aid_file_two.HTML"),
        PosixPath("dir_files/job_aid_file_one.mkd"),
        PosixPath("dir_files/job_aid_file_two.MKD"),
        PosixPath("dir_files/file_one.html"),
        PosixPath("dir_files/file_two.HTML"),
        PosixPath("dir_files/file_one.mkd"),
        PosixPath("dir_files/file_two.MKD")
    ])
    assert result1 == expected1
    result2 = sorted(dd.list_files(
        directory=directory,
        pattern_startswith=pattern_startswith
    ))
    expected2 = sorted([
        PosixPath("dir_files/job_aid_file_one.html"),
        PosixPath("dir_files/job_aid_file_two.HTML"),
        PosixPath("dir_files/job_aid_file_one.mkd"),
        PosixPath("dir_files/job_aid_file_two.MKD"),
        PosixPath("dir_files/file_one.html"),
        PosixPath("dir_files/file_two.HTML"),
        PosixPath("dir_files/file_one.mkd"),
        PosixPath("dir_files/file_two.MKD")
    ])
    assert result2 == expected2
    result3 = sorted(dd.list_files(
        directory=directory,
        pattern_extension=pattern_extension,
    ))
    expected3 = sorted([
        PosixPath("dir_files/job_aid_file_one.html"),
        PosixPath("dir_files/job_aid_file_two.HTML"),
        PosixPath("dir_files/job_aid_file_one.mkd"),
        PosixPath("dir_files/job_aid_file_two.MKD"),
        PosixPath("dir_files/file_one.html"),
        PosixPath("dir_files/file_two.HTML"),
        PosixPath("dir_files/file_one.mkd"),
        PosixPath("dir_files/file_two.MKD")
    ])
    assert result3 == expected3
    result4 = sorted(dd.list_files(
        directory=directory,
        pattern_startswith=pattern_startswith,
        pattern_extension=pattern_extension
    ))
    expected4 = sorted([
        PosixPath("dir_files/job_aid_file_one.html"),
        PosixPath("dir_files/job_aid_file_two.HTML"),
        PosixPath("dir_files/job_aid_file_one.mkd"),
        PosixPath("dir_files/job_aid_file_two.MKD"),
        PosixPath("dir_files/file_one.html"),
        PosixPath("dir_files/file_two.HTML"),
        PosixPath("dir_files/file_one.mkd"),
        PosixPath("dir_files/file_two.MKD")
    ])
    assert result4 == expected4


def test_byte_size():
    pass


def test_get_mtime():
    pass


def test_file_size():
    pass


def test_read_file():
    pass


def test_save_file():
    pass


def test_sort_rows():
    pass


def test_datetime():
    """Tests if specific columns are of datetime64[ns] type."""
    assert pd.api.types.is_datetime64_any_dtype(df["dates"])
    assert pd.api.types.is_datetime64_any_dtype(df["all_nat"])
    assert pd.api.types.is_datetime64_any_dtype(df["mixed"])


def test_integer_types():
    """Tests if integer-related columns are of the expected types."""
    assert pd.api.types.is_integer_dtype(df["integers"])
#     assert pd.api.types.is_integer_dtype(df["integers_int64"])
#     assert pd.api.types.is_integer_dtype(df["integers_uint8"])
#     assert pd.api.types.is_float_dtype(df["integers_float64"])
#     assert pd.api.types.is_float_dtype(df["floats"])


def test_float_types():
    """Tests if float columns are of the expected types."""
    assert pd.api.types.is_float_dtype(df["floats"])
    # assert pd.api.types.is_float_dtype(df["floats_float64"])
    # assert pd.api.types.is_float_dtype(df["floats_float32"])


def test_text_types():
    """Tests if text columns are of the expected types."""
    assert pd.api.types.is_object_dtype(df["text"])
    # assert pd.api.types.is_object_dtype(df["text_object"])
    # assert pd.api.types.is_string_dtype(df["text_string"])
